/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EARDeployableImpl.java 1406 2007-05-15 16:24:52Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployable.impl;

import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.easybeans.api.EZBArchive;
import org.objectweb.easybeans.api.EZBArchiveException;
import org.objectweb.easybeans.deployable.struct.ConnectorModuleImpl;
import org.objectweb.easybeans.deployable.struct.EARInfoImpl;
import org.objectweb.easybeans.deployable.struct.EJBModuleImpl;
import org.objectweb.easybeans.deployable.struct.JavaModuleImpl;
import org.objectweb.easybeans.deployable.struct.WebModuleImpl;
import org.objectweb.easybeans.deployer.api.CARDeployable;
import org.objectweb.easybeans.deployer.api.EARDeployable;
import org.objectweb.easybeans.deployer.api.EJB21Deployable;
import org.objectweb.easybeans.deployer.api.EJB3Deployable;
import org.objectweb.easybeans.deployer.api.EJBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployable;
import org.objectweb.easybeans.deployer.api.LibDeployable;
import org.objectweb.easybeans.deployer.api.RARDeployable;
import org.objectweb.easybeans.deployer.api.WARDeployable;
import org.objectweb.easybeans.deployer.api.struct.EARInfo;
import org.objectweb.easybeans.deployer.api.struct.WebModule;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.util.xml.DocumentParser;
import org.objectweb.easybeans.util.xml.DocumentParserException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Implementation for an EAR.
 * @author Florent Benoit
 */
public class EARDeployableImpl extends AbsDeployable<EARDeployable> implements EARDeployable {

    /**
     * Application.xml file entry in the EAR file.
     */
    private static final String APPLICATION_XML_ENTRY = "META-INF/application.xml";

    /**
     * Default library folder of an EAR.
     */
    private static final String DEFAULT_LIBRARY_FOLDER = "lib/";

    /**
     * Logger.
     */
    private JLog logger = JLogFactory.getLog(EARDeployableImpl.class);

    /**
     * Info of the application.xml file of this EAR.
     */
    private EARInfoImpl earInfo = null;

    /**
     * List of EJB 2.1 Deployables.
     */
    private List<EJB21Deployable> ejb21Deployables = null;

    /**
     * List of EJB 3 Deployables.
     */
    private List<EJB3Deployable> ejb3Deployables = null;

    /**
     * List of War Deployables.
     */
    private List<WARDeployable> warDeployables = null;

    /**
     * List of War Deployables.
     */
    private List<RARDeployable> rarDeployables = null;

    /**
     * List of War Deployables.
     */
    private List<CARDeployable> carDeployables = null;

    /**
     * List of Libraries Deployables.
     */
    private List<LibDeployable> libDeployables = null;

    /**
     * Defines and create a deployable for the given archive.
     * @param archive the given archive.
     */
    public EARDeployableImpl(final EZBArchive archive) {
        super(archive);
        this.ejb21Deployables = new ArrayList<EJB21Deployable>();
        this.ejb3Deployables = new ArrayList<EJB3Deployable>();
        this.warDeployables = new ArrayList<WARDeployable>();
        this.rarDeployables = new ArrayList<RARDeployable>();
        this.carDeployables = new ArrayList<CARDeployable>();
        this.libDeployables = new ArrayList<LibDeployable>();

        // analyze XML DD entry (if any)
        analyzeDD();
    }

    /**
     * Analyze the META-INF/application.xml entry if it is present.
     */
    private void analyzeDD() {
        // Compute URL to the META-INF/application.xml file.
        URL applicationXMLURL = null;
        try {
            applicationXMLURL = getArchive().getResource(APPLICATION_XML_ENTRY);
        } catch (EZBArchiveException e) {
            throw new IllegalStateException("Cannot get entry '" + APPLICATION_XML_ENTRY + "' on the archive '"
                    + getArchive().getName() + "'.", e);
        }

        // Create an XML representation of the file
        this.earInfo = new EARInfoImpl();

        // No XML file, just finish the treatment
        if (applicationXMLURL == null) {
            return;
        }

        // Get document without validation
        Document document = null;
        try {
            document = DocumentParser.getDocument(applicationXMLURL, false, null);
        } catch (DocumentParserException e) {
            throw new IllegalStateException("Cannot parse the url", e);
        }

        // Root element = <application>
        Element applicationRootElement = document.getDocumentElement();

        NodeList moduleList = applicationRootElement.getElementsByTagName("module");
        for (int i = 0; i < moduleList.getLength(); i++) {
            Element moduleElement = (Element) moduleList.item(i);

            // Now, we get the <module> element, need to get connector, ejb, web
            // or java elements

            // connector
            NodeList connectorList = moduleElement.getElementsByTagName("connector");
            for (int j = 0; j < connectorList.getLength(); j++) {
                String path = connectorList.item(j).getFirstChild().getNodeValue().trim();
                ConnectorModuleImpl connectorModule = new ConnectorModuleImpl();
                connectorModule.setPath(path);
                earInfo.addConnectorModule(connectorModule);
            }

            // EJB
            NodeList ejbList = moduleElement.getElementsByTagName("ejb");
            for (int j = 0; j < ejbList.getLength(); j++) {
                String path = ejbList.item(j).getFirstChild().getNodeValue().trim();
                EJBModuleImpl ejbModule = new EJBModuleImpl();
                ejbModule.setPath(path);
                earInfo.addEJBModule(ejbModule);
            }

            // Java
            NodeList javaList = moduleElement.getElementsByTagName("java");
            for (int j = 0; j < javaList.getLength(); j++) {
                String path = javaList.item(j).getFirstChild().getNodeValue().trim();
                JavaModuleImpl javaModule = new JavaModuleImpl();
                javaModule.setPath(path);
                earInfo.addJavaModule(javaModule);
            }

            // Web
            NodeList webList = moduleElement.getElementsByTagName("web");
            for (int j = 0; j < webList.getLength(); j++) {
                Element webElement = (Element) webList.item(j);

                // Need to get web-uri and context-root element
                String path = webElement.getElementsByTagName("web-uri").item(0).getFirstChild().getNodeValue().trim();
                String contextRoot = webElement.getElementsByTagName("context-root").item(0).getFirstChild().getNodeValue()
                        .trim();

                // Fix invalid context-root (no / at the beginning of the context, except for root context)
                if (contextRoot.startsWith("/") && !contextRoot.equals("/")) {
                    logger.debug("Context-Root ''{0}'' for module ''{1}'' contains invalid starting / in the name. Fixing it.",
                            contextRoot, path);
                    int c = 0;
                    while (contextRoot.charAt(c) == '/') {
                        c++;
                    }
                    contextRoot = contextRoot.substring(c);
                }

                WebModuleImpl webModule = new WebModuleImpl();
                webModule.setPath(path);
                webModule.setContextRoot(contextRoot);
                earInfo.addWebModule(webModule);

            }

        }

    }

    /**
     * @return the list of the EJB deployables for this EAR.
     */
    public List<EJBDeployable> getEJBDeployables() {
        // Sum of EJB 2.1 and EJB 3
        List<EJBDeployable> ejbDeployables = new ArrayList<EJBDeployable>();

        // 2.1
        for (EJB21Deployable deployable : this.ejb21Deployables) {
            ejbDeployables.add(deployable);
        }

        // 3
        for (EJB3Deployable deployable : this.ejb3Deployables) {
            ejbDeployables.add(deployable);
        }

        return ejbDeployables;

    }

    /**
     * @return the list of the EJB 2.1 deployables for this EAR.
     */
    public List<EJB21Deployable> getEJB21Deployables() {
        return ejb21Deployables;
    }

    /**
     * @return the list of the EJB 3 deployables for this EAR.
     */
    public List<EJB3Deployable> getEJB3Deployables() {
        return ejb3Deployables;
    }

    /**
     * @return the list of the War deployables for this EAR.
     */
    public List<WARDeployable> getWARDeployables() {
        return warDeployables;
    }

    /**
     * @return the list of the Rar deployables for this EAR.
     */
    public List<RARDeployable> getRARDeployables() {
        return rarDeployables;
    }

    /**
     * @return the list of the Car deployables for this EAR.
     */
    public List<CARDeployable> getCARDeployables() {
        return carDeployables;
    }

    /**
     * @return the list of the Library deployables for this EAR.
     */
    public List<LibDeployable> getLibDeployables() {
        return libDeployables;
    }

    /**
     * Add the given Deployable to this EAR deployable.
     * @param deployable the given deployable object to add
     */
    public void addDeployable(final EZBDeployable deployable) {

        // Change the type of the given deployable if it is a library for this EAR
        EZBDeployable ezbDeployable = checkLibDeployable(deployable);

        // Add in the right category depending on the interface

        if (ezbDeployable instanceof EJB21Deployable) {
            // EJB 2.1
            ejb21Deployables.add((EJB21Deployable) ezbDeployable);
        } else if (ezbDeployable instanceof EJB3Deployable) {
            // EJB 3
            ejb3Deployables.add((EJB3Deployable) ezbDeployable);
        } else if (ezbDeployable instanceof WARDeployable) {
            // WAR
            warDeployables.add((WARDeployable) ezbDeployable);

            // Try to set the right context-root for this module
            setContextRoot((WARDeployable) ezbDeployable);
        } else if (ezbDeployable instanceof RARDeployable) {
            // RAR
            rarDeployables.add((RARDeployable) ezbDeployable);
        } else if (ezbDeployable instanceof CARDeployable) {
            // CAR
            carDeployables.add((CARDeployable) ezbDeployable);
        } else if (ezbDeployable instanceof LibDeployable) {
            // Library
            libDeployables.add((LibDeployable) ezbDeployable);
        } else {
            // Not supported !
            throw new IllegalStateException("The EAR deployable archive doesn't support the include of archive with '"
                    + ezbDeployable.getClass() + "' class object, deployable ='" + ezbDeployable + "'.");
        }
    }

    /**
     * If it is a library, change the type of the deployable.
     * @param deployable the module of this EAR
     * @return the new deployable if it is a library module.
     */
    protected EZBDeployable checkLibDeployable(final EZBDeployable deployable) {

        // Get name within the EAR
        String relativeModuleName = getRelativeNameFromDeployable(deployable);

        // It is a library
        if (relativeModuleName.startsWith(DEFAULT_LIBRARY_FOLDER)) {
            return new LibDeployableImpl(deployable.getArchive());
        }

        return deployable;
    }

    /**
     * Return the relative name of the deployable from the EAR deployable.
     * @param deployable the module of the EAR
     * @return the relative name
     */
    protected String getRelativeNameFromDeployable(final EZBDeployable deployable) {
        // get URL of the EAR file
        URL earURL = null;
        try {
            earURL = getArchive().getURL();
        } catch (EZBArchiveException e) {
            throw new IllegalStateException("Cannot get the URL for the EAR file '" + this + "'.", e);
        }

        // get the URL of the deployable
        URL deployableURL = null;
        try {
            deployableURL = deployable.getArchive().getURL();
        } catch (EZBArchiveException e) {
            throw new IllegalStateException("Cannot get the URL for the deployable '" + deployable + "'.", e);
        }

        return deployableURL.toExternalForm().substring(earURL.toExternalForm().length());
    }

    /**
     * Set the name of the context-root for the given WAR by checking the EAR
     * information (if any).
     * @param warDeployable the given war to analyze
     */
    protected void setContextRoot(final WARDeployable warDeployable) {

        // Extract the relative name of the war from the EAR
        String relativeWarFileName = getRelativeNameFromDeployable(warDeployable);

        // now, try to find a matching value from the parsed application.xml
        // file
        boolean found = false;
        for (WebModule webModule : earInfo.getWebs()) {
            if (relativeWarFileName.equals(webModule.getPath())) {
                // Found the correct entry
                String ctxRoot = webModule.getContextRoot();
                warDeployable.setContextRoot(ctxRoot);
                if (logger.isDebugEnabled()) {
                    logger.debug("Context root for War deployable ''{0}'' was found in the EAR "
                            + "{1} file, the value is ''{2}''", warDeployable, APPLICATION_XML_ENTRY, ctxRoot);
                }
                found = true;
                break;
            }
        }

        // Not found, apply a default context root
        if (!found) {

            // get the URL of the deployable
            URL warURL = null;
            try {
                warURL = warDeployable.getArchive().getURL();
            } catch (EZBArchiveException e) {
                throw new IllegalStateException("Cannot get the URL for the deployable '" + warDeployable + "'.", e);
            }

            // context root
            String path = warURL.getPath();
            int sep = path.lastIndexOf("/");
            int dot = path.lastIndexOf(".");
            String ctxRoot = path.substring(sep + 1, dot);
            warDeployable.setContextRoot(ctxRoot);
            if (logger.isDebugEnabled()) {
                logger.debug("The entry for War deployable ''{0}'' was not found in the EAR "
                        + "application.xml file, using a default context-root value ''{1}''", warDeployable, ctxRoot);
            }

        }

    }

    /**
     * @return EAR information (of the application.xml file).
     */
    public EARInfo getEARInfo() {
        return earInfo;
    }
}
