/**
 * EasyBeans
 * Copyright (C) 2006-2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: XMLConfiguration.java 1387 2007-05-15 07:49:25Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.xmlconfig;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.util.xml.DocumentParser;
import org.objectweb.easybeans.util.xml.DocumentParserException;
import org.objectweb.easybeans.xmlconfig.mapping.AttributeMapping;
import org.objectweb.easybeans.xmlconfig.mapping.ClassMapping;
import org.objectweb.easybeans.xmlconfig.mapping.XMLMapping;
import org.objectweb.easybeans.xmlconfig.mapping.XMLMappingBuilder;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * XML configuration class that configure a given object with an xml
 * configuration file. <br>
 * Additional information will be searched on the namespace with the resource named easybeans-mapping.xml.
 * @author Florent Benoit
 */
public class XMLConfiguration {

    /**
     * Http:// namespace.
     */
    private static final String HTTP_NAMESPACE = "http://";

    /**
     * By default, no validation.
     */
    private static final boolean VALIDATING_OFF = false;

    /**
     * Per-Package mapping file name.
     */
    private static final String MAPPING_FILENAME = "easybeans-mapping.xml";

    /**
     * Logger.
     */
    private JLog logger = JLogFactory.getLog(XMLConfiguration.class);

    /**
     * Validating when parsing xml ?
     */
    private boolean validating = VALIDATING_OFF;

    /**
     * URL of the configuration file.
     */
    private URL xmlConfigurationURL;

    /**
     * Mapping between namespace and XMLMapping object.
     */
    private Map<String, XMLMapping> mappings = null;

    /**
     * Link between name and the objects.
     */
    private HashMap<String, Object> configuredElements = null;

    /**
     * Build an xml configuration with the given xml configuration file.
     * @param xmlConfigurationURL the given xml configuration file
     */
    public XMLConfiguration(final URL xmlConfigurationURL) {
        this.xmlConfigurationURL = xmlConfigurationURL;
        this.mappings = new HashMap<String, XMLMapping>();
        configuredElements = new HashMap<String, Object>();
    }

    /**
     * Configure the given object with the configuration that was analyzed.
     * @param object the object to configure.
     * @throws XMLConfigurationException if configuration fails
     */
    public void configure(final Object object) throws XMLConfigurationException {
        Element rootElement = analyzeXmlFile();
        configure(object, rootElement);
    }

    /**
     * Gets the root element of the XML parsed file.
     * @return the element object
     * @throws XMLConfigurationException if analyze fails.
     */
    private Element analyzeXmlFile() throws XMLConfigurationException {
        // Parse the XML file and build all configuration process.
        Document xmlConfigurationDocument = null;
        try {
            xmlConfigurationDocument = DocumentParser.getDocument(xmlConfigurationURL, validating, null);
        } catch (DocumentParserException e) {
            throw new XMLConfigurationException("Cannot get a document on the given url '" + xmlConfigurationURL + "'.", e);
        }

        // Get the root element
        Element rootElement = xmlConfigurationDocument.getDocumentElement();

        // get namespace of the element
        String nameSpace = rootElement.getNamespaceURI();
        // and the name
        // String nodeName = rootElement.getNodeName();

        // init mapping for the element
        getXMLMapping(nameSpace);

        return rootElement;

    }

    /**
     * Configure the given object by looking at the element (of the
     * configuration).
     * @param object the given object to configure.
     * @param element the element that contains configuration.
     * @throws XMLConfigurationException if configuration fails
     */
    public void configure(final Object object, final Element element) throws XMLConfigurationException {

        // Get all children of the given element
        NodeList nodeList = element.getChildNodes();

        // For each element, create object and configure it
        int length = nodeList.getLength();
        for (int i = 0; i < length; i++) {
            Node node = nodeList.item(i);

            // Get an element, create an instance of the element
            if (Node.ELEMENT_NODE == node.getNodeType()) {
                ClassMapping classMapping = getClassMapping(node.getNamespaceURI(), node.getNodeName());

                // Build an object for the given class mapping
                Object newObject = newInstance(node, classMapping);

                // Try to see if there is a value for the current object
                Method getterMethod = findSingleGetterMethod(object, newObject);
                if (getterMethod != null) {
                    // Call the getter method
                    logger.debug("Calling method {0} on object {1}", getterMethod, object);
                    Object getterObject = null;
                    try {
                        getterObject = getterMethod.invoke(object);
                    } catch (IllegalArgumentException e) {
                        throw new XMLConfigurationException("Cannot call method '" + getterMethod + "' on object '" + object
                                + "'.", e);
                    } catch (IllegalAccessException e) {
                        throw new XMLConfigurationException("Cannot call method '" + getterMethod + "' on object '" + object
                                + "'.", e);
                    } catch (InvocationTargetException e) {
                        throw new XMLConfigurationException("Cannot call method '" + getterMethod + "' on object '" + object
                                + "'.", e);
                    }

                    // find an object ?
                    if (getterObject != null) {
                        // yes, then use the existing object
                        newObject = getterObject;
                    }
                }

                // Set attributes
                NamedNodeMap attributes = ((Element) node).getAttributes();
                int attributesLength = attributes.getLength();
                for (int a = 0; a < attributesLength; a++) {
                    Node nodeAttribute = attributes.item(a);
                    if (nodeAttribute.getNamespaceURI() == null) {
                        String propertyName = nodeAttribute.getNodeName();
                        String propertyValue = nodeAttribute.getNodeValue();
                        setAttribute(newObject, propertyName, propertyValue, classMapping);
                    }
                }
                // add or set element on the given object
                addSetElement(object, newObject);

                // Store element
                configuredElements.put(node.getNodeName(), newObject);

                // configure sub element of the new object.
                configure(newObject, (Element) node);
            }
        }
    }

    /**
     * Gets the Mapping for the given nameSpaceURI and the given classname.
     * @param nameSpaceURI the namespace of the xml element
     * @param name the name of the class
     * @return the class mapping object
     * @throws XMLConfigurationException if no configuration is found
     */
    private ClassMapping getClassMapping(final String nameSpaceURI, final String name) throws XMLConfigurationException {
        // Get classname with the namespace and alias
        XMLMapping xmlMapping = mappings.get(nameSpaceURI);
        if (xmlMapping == null) {
            // Not yet encountered namespace. Try to get a mapping for it.
            xmlMapping = getXMLMapping(nameSpaceURI);
            if (xmlMapping == null) {
                // Still not found.
                throw new XMLConfigurationException("No mapping found for namespace '" + nameSpaceURI + "'.");
            }
        }

        // alias present ?
        ClassMapping classMapping = xmlMapping.getClassMapping(name);

        // no mapping
        if (classMapping == null) {
            throw new XMLConfigurationException("No class found for element '" + name + "' withing xmlmapping '" + xmlMapping
                    + "'.");
        }
        return classMapping;
    }

    /**
     * Build class instance for the given node.
     * @param node the element of the xml
     * @param classMapping the mapping between node name and class name
     * @return the instance of the object
     * @throws XMLConfigurationException if no new instance can be done.
     */
    private Object newInstance(final Node node, final ClassMapping classMapping) throws XMLConfigurationException {
        // get classname
        String className = classMapping.getName();

        // load class
        Class clazz;
        try {
            clazz = Thread.currentThread().getContextClassLoader().loadClass(className);
        } catch (ClassNotFoundException e) {
            throw new XMLConfigurationException("Cannot load the class '" + className + "'.", e);
        }
        logger.debug("Building instance of the class {0}", className);
        // build a new instance
        try {
            return clazz.newInstance();
        } catch (InstantiationException e) {
            throw new XMLConfigurationException("Cannot build an instance of class '" + className + "'.", e);
        } catch (IllegalAccessException e) {
            throw new XMLConfigurationException("Cannot build an instance of class '" + className + "'.", e);
        }
    }

    /**
     * Set attribute on the given object.
     * @param configuringObject the object to configure
     * @param propertyName the name of the property (for the setter)
     * @param propertyValue the value of the property
     * @param classMapping the mapping (if there are attributes aliasing)
     * @throws XMLConfigurationException if attribute cannot be set.
     */
    private void setAttribute(final Object configuringObject, final String propertyName, final String propertyValue,
            final ClassMapping classMapping) throws XMLConfigurationException {

        // get class of the configuringObject
        Class configuringClass = configuringObject.getClass();

        // Get methods
        Method[] methods = configuringClass.getMethods();

        String updatedPropertyName = propertyName;

        // alias ?
        AttributeMapping attributeMapping = classMapping.getAttributeMapping(propertyName);
        if (attributeMapping != null) {
            String name = attributeMapping.getName();
            if (name != null) {
                updatedPropertyName = name;
            }
        }

        String setterName = "set" + updatedPropertyName.substring(0, 1).toUpperCase() + updatedPropertyName.substring(1);

        Object paramObject = null;
        // Search setter
        Method foundMethod = null;
        if (methods != null) {
            for (Method m : methods) {
                // set or add ?
                if (m.getName().startsWith(setterName)) {
                    // only one arg ?
                    Class[] parameters = m.getParameterTypes();
                    if (parameters.length == 1) {
                        foundMethod = m;
                        Class parameterClass = parameters[0];

                        // Special case of reference
                        if (propertyValue.startsWith("#") && propertyValue.length() >= 2) {
                            paramObject = configuredElements.get(propertyValue.substring(1));
                        } else if (String.class.equals(parameterClass)) {
                            paramObject = propertyValue;
                        } else if (Integer.TYPE.equals(parameterClass)) {
                            paramObject = new Integer(propertyValue);
                        } else if (Boolean.TYPE.equals(parameterClass)) {
                            paramObject = Boolean.valueOf(propertyValue);
                        }
                        break;
                    }
                }
            }
        }

        // build
        if (foundMethod != null && paramObject != null) {
            logger.debug("Calling {0} on object {1} with value {2}", setterName, configuringObject, paramObject);
            try {
                foundMethod.invoke(configuringObject, paramObject);
            } catch (IllegalArgumentException e) {
                throw new XMLConfigurationException("Cannot call method '" + foundMethod + "' on object '" + configuringObject
                        + "' with parameter '" + paramObject + "'.", e);
            } catch (IllegalAccessException e) {
                throw new XMLConfigurationException("Cannot call method '" + foundMethod + "' on object '" + configuringObject
                        + "' with parameter '" + paramObject + "'.", e);
            } catch (InvocationTargetException e) {
                throw new XMLConfigurationException("Cannot call method '" + foundMethod + "' on object '" + configuringObject
                        + "' with parameter '" + paramObject + "'.", e);
            }
        } else {
            throw new XMLConfigurationException("No setter method found for parameter '" + propertyName + "' on class '"
                    + configuringClass + "'.");
        }
    }

    /**
     * Finds the add(Class c) or set(Class c) methods.<br>
     * The parameter needs to be a single entry and not a list.
     * @param configuringObject the object on which call this setter method.
     * @param paramObject the parameter object
     * @return the found method if any.
     */
    private Method findSingleSetterMethod(final Object configuringObject, final Object paramObject) {
        Method setterMethod = null;

        // get class of the configuringObject
        Class configuringClass = configuringObject.getClass();

        // Get methods of the class
        Method[] methods = configuringClass.getMethods();

        // class of the new object
        Class parameterClass = paramObject.getClass();
        String className = parameterClass.getSimpleName();

        // Search setter
        if (methods != null) {
            for (Method m : methods) {
                // set or add ?
                if (m.getName().startsWith("set" + className) || m.getName().startsWith("add" + className)) {
                    // only one arg ?
                    Class[] parameters = m.getParameterTypes();
                    if (parameters.length == 1) {
                        // valid parameter
                        if (parameters[0].equals(parameterClass)) {
                            setterMethod = m;
                            break;
                        }
                    }
                }
            }
        }
        return setterMethod;
    }

    /**
     * Finds the get(Class c) method.<br>
     * The parameter needs to be a single entry and not a list.
     * @param configuringObject the object on which call this getter method.
     * @param paramObject the parameter object
     * @return the found method if any.
     */
    private Method findSingleGetterMethod(final Object configuringObject, final Object paramObject) {
        Method getterMethod = null;

        // get class of the configuringObject
        Class configuringClass = configuringObject.getClass();

        // Get methods of the class
        Method[] methods = configuringClass.getMethods();

        // class of the new object
        Class parameterClass = paramObject.getClass();
        String className = parameterClass.getSimpleName();

        // Search setter
        if (methods != null) {
            for (Method m : methods) {
                // Gettermethod ?
                if (m.getName().startsWith("get" + className)) {
                    // no Arg ?
                    Class[] parameters = m.getParameterTypes();
                    if (parameters.length == 0) {
                        // valid parameter and return type ?
                        Class returnType = m.getReturnType();
                        if (returnType != null && returnType.equals(paramObject.getClass())) {
                            getterMethod = m;
                            break;
                        }
                    }
                }
            }
        }
        return getterMethod;
    }

    /**
     * Finds the get(List&lt;Class&gt;) and set(List&lt;Class&gt;) methods. The
     * parameter needs to be a list with generics.
     * @param configuringObject the object on which call this setter method.
     * @param paramObject the parameter object
     * @return the getter and setter method with List type.
     */
    private Method[] findListGetterSetterMethod(final Object configuringObject, final Object paramObject) {

        // get class of the configuringObject
        Class configuringClass = configuringObject.getClass();

        // Get methods of the class
        Method[] methods = configuringClass.getMethods();

        // class of the new object
        Class parameterClass = paramObject.getClass();

        // Search with class of object
        String className = parameterClass.getSimpleName();
        // Suffix name (many items)
        // (special case for y)
        String argName = null;
        if (className.endsWith("y")) {
            argName = className.substring(0, className.length() - 1) + "ies";
        } else {
            argName = className + "s";
        }

        Method[] searchMethods = searchListGetterSetterMethod(methods, argName, parameterClass);

        return searchMethods;
    }

    /**
     * Search getter and setter method in the given set of methods. The
     * getter/setter are used for a List parameter/return type.
     * @param methods the set of methods to search in
     * @param argName the name of the argument
     * @param parameterClass the parameter class (for the generics on the List)
     * @return an array of methods, array[0] = getter, array[1] = setter.
     */
    private Method[] searchListGetterSetterMethod(final Method[] methods, final String argName, final Class parameterClass) {
        Method[] returnedMethods = null;

        Method getterMethod = null;
        Method setterMethod = null;
        // Search getter and setter
        if (methods != null) {
            for (Method m : methods) {
                // setter
                if (m.getName().startsWith("set" + argName)) {
                    // only one arg ?
                    Class[] parameters = m.getParameterTypes();

                    // setter
                    if (parameters.length == 1) {
                        // ensure it is a List type
                        if (parameters[0].equals(List.class)) {
                            Type[] types = m.getGenericParameterTypes();
                            // look only the first type if any
                            if (types.length == 1) {
                                if (types[0] instanceof ParameterizedType) {
                                    ParameterizedType parameterizedType = (ParameterizedType) types[0];
                                    Type[] typeArguments = parameterizedType.getActualTypeArguments();
                                    if (typeArguments.length == 1) {
                                        if (typeArguments[0].equals(parameterClass)) {
                                            // found it !
                                            setterMethod = m;
                                        }
                                    }

                                }
                            }
                        }
                    }
                } else if (m.getName().startsWith("get" + argName)) {
                    Class returnTypeClass = m.getReturnType();
                    if (returnTypeClass != null && returnTypeClass.equals(List.class)) {

                        // no parameter
                        Class[] parameters = m.getParameterTypes();
                        if (parameters.length == 0) {
                            Type type = m.getGenericReturnType();
                            if (type instanceof ParameterizedType) {
                                ParameterizedType parameterizedType = (ParameterizedType) type;
                                Type[] typeArguments = parameterizedType.getActualTypeArguments();
                                if (typeArguments.length == 1) {
                                    if (typeArguments[0].equals(parameterClass)) {
                                        // found it !
                                        getterMethod = m;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        // try with interfaces of the current class
        Class[] interfaces = parameterClass.getInterfaces();
        if (interfaces != null) {
            for (Class itf : interfaces) {
                String itfName = itf.getSimpleName();
                String argItfName = itfName + "s";
                returnedMethods = searchListGetterSetterMethod(methods, argItfName, itf);
                if (returnedMethods != null) {
                    break;
                }
            }
        }

        if (getterMethod != null && setterMethod != null) {
            returnedMethods = new Method[2];
            returnedMethods[0] = getterMethod;
            returnedMethods[1] = setterMethod;
        }
        return returnedMethods;
    }

    /**
     * Add or set element on the given object.
     * @param configuringObject the object to configure
     * @param newObject the object to add/set on the configuringObject object
     * @throws XMLConfigurationException if newObject cannot be set.
     */
    @SuppressWarnings("unchecked")
    private void addSetElement(final Object configuringObject, final Object newObject) throws XMLConfigurationException {
        Method foundMethod = findSingleSetterMethod(configuringObject, newObject);
        // single setter ?
        if (foundMethod != null) {
            logger.debug("Calling method {0} on object {1} with value {2}", foundMethod, configuringObject, newObject);
            try {
                foundMethod.invoke(configuringObject, newObject);
            } catch (IllegalArgumentException e) {
                throw new XMLConfigurationException("Unable to invoke the method '" + foundMethod + "' on '" + configuringObject
                        + "' with arg '" + newObject + "'.", e);
            } catch (IllegalAccessException e) {
                throw new XMLConfigurationException("Unable to invoke the method '" + foundMethod + "' on '" + configuringObject
                        + "' with arg '" + newObject + "'.", e);
            } catch (InvocationTargetException e) {
                throw new XMLConfigurationException("Unable to invoke the method '" + foundMethod + "' on '" + configuringObject
                        + "' with arg '" + newObject + "'.", e);
            }
        } else {
            // try a list getter/setter
            Method[] getterAndSetter = findListGetterSetterMethod(configuringObject, newObject);
            if (getterAndSetter != null) {
                Method getterMethod = getterAndSetter[0];
                Method setterMethod = getterAndSetter[1];

                // Get current list
                List currentList = null;
                try {
                    logger.debug("Calling method {0} on object {1}", getterMethod, configuringObject);
                    currentList = (List) getterMethod.invoke(configuringObject);
                } catch (IllegalArgumentException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + getterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (IllegalAccessException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + getterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (InvocationTargetException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + getterMethod + "' on '"
                            + configuringObject + "'.", e);
                }

                // Null list, create one
                if (currentList == null) {
                    currentList = new ArrayList();
                }

                // add element
                logger.debug("Adding element  {0} to the list {1}", newObject, currentList);
                currentList.add(newObject);

                // set list
                try {
                    logger.debug("Calling method {0} on object {1} with value {2}", setterMethod, configuringObject, currentList);
                    setterMethod.invoke(configuringObject, currentList);
                } catch (IllegalArgumentException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + setterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (IllegalAccessException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + setterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (InvocationTargetException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + setterMethod + "' on '"
                            + configuringObject + "'.", e);
                }

            } else {
                throw new XMLConfigurationException("No setter method found for parameter '" + configuringObject.getClass()
                        + "' on class '" + newObject.getClass() + "'.");
            }
        }
    }

    /**
     * Gets the XML mapping for the given namespace and the node name.
     * @param nameSpace current XML namespace.
     * @return the XML mapping object.
     * @throws XMLConfigurationException if xml mapping is not found.
     */
    private XMLMapping getXMLMapping(final String nameSpace) throws XMLConfigurationException {
        if (nameSpace == null || "".equals(nameSpace)) {
            throw new XMLConfigurationException("Empty default namespace not supported.");
        }

        // already one ?
        if (mappings.containsKey(nameSpace)) {
            return mappings.get(nameSpace);
        }

        XMLMapping xmlMapping = null;
        // With the namespace, get the package name for searching the
        // <nodeName>-mapping.xml file
        if (nameSpace != null) {
            if (nameSpace.startsWith(HTTP_NAMESPACE)) {
                // get package
                String packageName = nameSpace.substring(HTTP_NAMESPACE.length());
                // Compute node name and suffix
                String resourceName = packageName.replaceAll("\\.", "/") + "/" + MAPPING_FILENAME;

                URL mappingURL = Thread.currentThread().getContextClassLoader().getResource(resourceName);
                // found url
                if (mappingURL != null) {
                    xmlMapping = createXMLMapping(nameSpace, mappingURL);
                }
            } else {
                logger.warn("Namespace found ''{0}}'' but this is not an HTTP namespace", nameSpace);
            }
        }
        // even null namespace will have a default xml mapping
        mappings.put(nameSpace, xmlMapping);
        return xmlMapping;

    }

    /**
     * Create the XML mapping object for the given namespace with the given XML
     * mapping file.
     * @param nameSpace the namespace of the mapping
     * @param mappingURL the URL to the XML mapping file
     * @return an instance of XML mapping
     * @throws XMLConfigurationException if no XML mapping have been created.
     */
    private XMLMapping createXMLMapping(final String nameSpace, final URL mappingURL) throws XMLConfigurationException {
        XMLMappingBuilder xmlMappingBuilder = new XMLMappingBuilder(mappingURL);
        xmlMappingBuilder.build();
        return xmlMappingBuilder.getXmlMapping();
    }
}
