/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbsDeployer.java 1417 2007-05-16 16:12:07Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployer;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.objectweb.easybeans.api.EZBArchive;
import org.objectweb.easybeans.api.EZBArchiveException;
import org.objectweb.easybeans.api.EZBContainer;
import org.objectweb.easybeans.api.EZBContainerException;
import org.objectweb.easybeans.api.EZBServer;
import org.objectweb.easybeans.deployable.DeployableHelper;
import org.objectweb.easybeans.deployable.DeployableHelperException;
import org.objectweb.easybeans.deployable.archive.ArchiveManager;
import org.objectweb.easybeans.deployable.impl.EARDeployableImpl;
import org.objectweb.easybeans.deployer.api.EARDeployable;
import org.objectweb.easybeans.deployer.api.EJB3Deployable;
import org.objectweb.easybeans.deployer.api.EJBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployer;
import org.objectweb.easybeans.deployer.api.EZBDeployerException;
import org.objectweb.easybeans.deployer.api.LibDeployable;
import org.objectweb.easybeans.loader.EasyBeansClassLoader;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.persistence.PersistenceUnitManager;
import org.objectweb.easybeans.persistence.xml.PersistenceXmlFileAnalyzer;
import org.objectweb.easybeans.persistence.xml.PersistenceXmlFileAnalyzerException;
import org.objectweb.easybeans.util.files.FileUtils;
import org.objectweb.easybeans.util.files.FileUtilsException;

/**
 * Abstract class that defines common methods for deployer.
 * @author Florent Benoit
 */
public abstract class AbsDeployer implements EZBDeployer {

    /**
     * Folder to create in tmp folder.
     */
    public static final String DEFAULT_FOLDER = "EasyBeans-Deployer";

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(AbsDeployer.class);

    /**
     * Embedded server linked to this deployer.
     */
    private EZBServer embedded = null;

    /**
     * @return the embedded instance used by this server.
     */
    public EZBServer getEmbedded() {
        return embedded;
    }

    /**
     * Receive Embedded instance for this deployer.
     * @param embedded the given instance of the embedded server.
     */
    public void setEmbedded(final EZBServer embedded) {
        this.embedded = embedded;
    }

    /**
     * Build and return a classloader for the given EAR.
     * @param earDeployable the given EAR
     * @return a classloader
     * @throws EZBDeployerException if the classloader cannot be built
     */
    protected ClassLoader getClassLoaderForEAR(final EARDeployable earDeployable) throws EZBDeployerException {
        // Get EJBs of this EAR
        List<EJBDeployable> ejbs = earDeployable.getEJBDeployables();

        // Get libraries of this EAR
        List<LibDeployable> libs = earDeployable.getLibDeployables();

        // Create array of URLs with EJBs + Libraries
        List<URL> urls = new ArrayList<URL>();
        for (EJBDeployable ejb : ejbs) {
            try {
                urls.add(ejb.getArchive().getURL());
            } catch (EZBArchiveException e) {
                throw new EZBDeployerException("Cannot get the URL for the Archive '" + ejb.getArchive() + "'.", e);
            }
        }
        for (LibDeployable lib : libs) {
            try {
                urls.add(lib.getArchive().getURL());
            } catch (EZBArchiveException e) {
                throw new EZBDeployerException("Cannot get the URL for the Archive '" + lib.getArchive() + "'.", e);

            }
        }

        // Create classloader with these URLs
        URL[] arrayURLs = urls.toArray(new URL[urls.size()]);
        return new EasyBeansClassLoader(arrayURLs, Thread.currentThread().getContextClassLoader());
    }

    /**
     * Gets the persistence unit manager for the given EAR and classloader.
     * @param earDeployable the ear deployable
     * @param appClassLoader the classloader used as deployable
     * @return the given persistence unit manager
     */
    protected PersistenceUnitManager getPersistenceUnitManager(final EARDeployable earDeployable,
            final ClassLoader appClassLoader) {
        // Analyze libraries to detect persistence archive (only once for now
        // and for all libraries)
        // Get libraries of this EAR
        List<LibDeployable> libs = earDeployable.getLibDeployables();
        PersistenceUnitManager persistenceUnitManager = null;
        for (LibDeployable lib : libs) {
            PersistenceUnitManager builtPersistenceUnitManager = null;
            try {
                builtPersistenceUnitManager = PersistenceXmlFileAnalyzer.analyzePersistenceXmlFile(lib.getArchive(),
                        appClassLoader);
            } catch (PersistenceXmlFileAnalyzerException e) {
                throw new IllegalStateException("Failure when analyzing the persistence.xml file", e);
            }
            if (persistenceUnitManager != null) {
                throw new IllegalStateException("There is an existing persistence.xml in another library of this EAR."
                        + " This is not supported.");
            }
            persistenceUnitManager = builtPersistenceUnitManager;
        }
        return persistenceUnitManager;
    }

    /**
     * Deploy an EJB (called by the deploy method).
     * @param ejbDeployable a given EJB deployable
     * @throws EZBDeployerException if the deployment is not done.
     */
    protected void deployEJB(final EJBDeployable ejbDeployable) throws EZBDeployerException {
        logger.info("Deploying {0}", ejbDeployable);
        EZBContainer container = getEmbedded().createContainer(ejbDeployable.getArchive());
        try {
            container.start();
        } catch (EZBContainerException e) {
            getEmbedded().removeContainer(container);
            throw new EZBDeployerException("Cannot deploy the given EJB '" + ejbDeployable + "'.", e);
        }
    }

    /**
     * Build an instance of the given class.
     * @param clazz the class to instantiate
     * @return a new object
     * @throws EZBDeployerException if the class can't be loaded
     */
    protected static Object newInstance(final Class clazz) throws EZBDeployerException {
        try {
            return clazz.newInstance();
        } catch (InstantiationException e) {
            throw new EZBDeployerException("Cannot make an instance of the class '" + clazz + "'.", e);
        } catch (IllegalAccessException e) {
           throw new EZBDeployerException("Cannot make an instance of the class '" + clazz + "'.", e);
        }
    }

    /**
     * Build an instance by using the given constructor and given parameters.
     * @param constructor the constructor to use
     * @param parameters the parameters of the given constructor
     * @return a new object
     * @throws EZBDeployerException if the class can't be loaded
     */
    protected static Object newInstance(final Constructor constructor, final Object... parameters)
            throws EZBDeployerException {
        try {
            return constructor.newInstance(parameters);
        } catch (IllegalArgumentException e) {
            throw new EZBDeployerException("Cannot create a classloader with constructor '" + constructor + "'", e);
        } catch (InstantiationException e) {
            throw new EZBDeployerException("Cannot create a classloader with constructor '" + constructor + "'", e);
        } catch (IllegalAccessException e) {
            throw new EZBDeployerException("Cannot create a classloader with constructor '" + constructor + "'", e);
        } catch (InvocationTargetException e) {
            throw new EZBDeployerException("Cannot create a classloader with constructor '" + constructor + "'", e);
        }
    }



    /**
     * Load the given class with its given classname.
     * @param className the name of the class to load
     * @return the class object
     * @throws EZBDeployerException if the class can't be loaded
     */
    protected static Class loadClass(final String className) throws EZBDeployerException {
        return loadClass(className, null);
    }

    /**
     * Load the given class with its given classname.
     * @param className the name of the class to load
     * @param classLoader the given classloader (or null to use thread context classloader)
     * @return the class object
     * @throws EZBDeployerException if the class can't be loaded
     */
    protected static Class loadClass(final String className, final ClassLoader classLoader) throws EZBDeployerException {
        // Load the classr
        Class clazz = null;
        try {
            if (classLoader != null) {
                clazz = classLoader.loadClass(className);
            } else {
                clazz = Thread.currentThread().getContextClassLoader().loadClass(className);
            }
        } catch (ClassNotFoundException e) {
            throw new EZBDeployerException("Cannot load the class '" + className + "'", e);
        }
        return clazz;
    }

    /**
     * Invoke the given method on the given object (null for static method) and
     * with the given args.
     * @param method the method to invoke
     * @param object the object on which the method is invoked
     * @param args the arguments of the method
     * @return the result of the invocation
     * @throws EZBDeployerException if the method is not invoked
     */
    protected static Object invoke(final Method method, final Object object, final Object... args) throws EZBDeployerException {
        try {
            return method.invoke(object, args);
        } catch (IllegalArgumentException e) {
            throw new EZBDeployerException("Cannot invoke the method '" + method + "'", e);
        } catch (IllegalAccessException e) {
            throw new EZBDeployerException("Cannot invoke the method '" + method + "'", e);
        } catch (InvocationTargetException e) {
            throw new EZBDeployerException("Cannot invoke the method '" + method + "'", e);
        }
    }

    /**
     * Get the method on the given class with the given method name and the
     * given parameters.
     * @param clazz the class on which search the method
     * @param methodName the name of the method to search
     * @param parameters the class parameters of the method that is searched
     * @return the method found
     * @throws EZBDeployerException if the method is not found
     */
    protected static Method getMethod(final Class clazz, final String methodName, final Class... parameters)
        throws EZBDeployerException {
        try {
            return clazz.getMethod(methodName, parameters);
        } catch (SecurityException e) {
            throw new EZBDeployerException("Cannot get the Method '" + methodName + "' on the '" + clazz + "' class.", e);
        } catch (NoSuchMethodException e) {
            throw new EZBDeployerException("Cannot get the Method '" + methodName + "' on the '" + clazz + "' class.", e);
        }
    }


    /**
     * Undeploy EJB3s of an EAR (called by the undeploy method).
     * @param earDeployable a given EAR deployable
     * @throws EZBDeployerException if the deployment is not done.
     */
    protected void undeployEJB3FromEAR(final EARDeployable earDeployable) throws EZBDeployerException {
        // From which deployable get the containers deployed
        EARDeployable workingDeployable = earDeployable;

        // Check if this archive has been unpacked ?
        EARDeployable unpackedDeployable = earDeployable.getUnpackedDeployable();
        if (unpackedDeployable != null) {
            workingDeployable = unpackedDeployable;
        }

        // Get Containers of this deployable
        List<EZBContainer> containers = new ArrayList<EZBContainer>();
        for (EJB3Deployable ejb3 : workingDeployable.getEJB3Deployables()) {
            EZBContainer container = getEmbedded().findContainer(ejb3.getArchive());
            // not found
            if (container == null) {
                logger.warn("No container found for the archive ''{0}'', creation has maybe failed", ejb3.getArchive());
                continue;
            }
            // found, add it
            containers.add(container);
        }

        // Remove all these containers
        for (EZBContainer container : containers) {
            // stop it
            container.stop();

            // remove it
            getEmbedded().removeContainer(container);
        }
    }

    /**
     * Unpack the given archive in a temp folder, then build a local EARDeployable and fill it with submodules and then return it.
     * @param earDeployable the archive to unpack.
     * @return a new deployable (which is unpacked)
     * @throws EZBDeployerException if the EAR can't be unpacked
     */
    protected EARDeployable unpackEARDeployable(final EARDeployable earDeployable) throws EZBDeployerException {
        EARDeployableImpl unpackedDeployable = null;

        // Build a temp directory (which is root of all unpack)
        File rootFolder = new File(System.getProperty("java.io.tmpdir") + File.separator + DEFAULT_FOLDER
                + "-" + System.getProperty("user.name", "default"));
        rootFolder.mkdirs();

        // Create EAR directory in this folder (if it is not existing) for unpacking EAR
        File earFolder =  new File(rootFolder, "EAR");
        earFolder.mkdirs();


        // Now, create a directory with the name of the given EAR
        URL earURL = null;
        String pathURL = null;
        try {
            earURL =  earDeployable.getArchive().getURL();
        } catch (EZBArchiveException e) {
            throw new EZBDeployerException("Cannot get the URL for the deployable '" + earDeployable + "'.", e);
        }

        // extract filename from the URL for creating the directory
        pathURL = earURL.getPath();
        String fileName = pathURL.substring(pathURL.lastIndexOf("/") + 1);
        File earUnpackedFolder = new File(earFolder, fileName);
        earUnpackedFolder.mkdir();

        // Create new deployable
        EZBArchive builtArchive = ArchiveManager.getInstance().getArchive(earUnpackedFolder);

        // Now, do a dump of each resource of the given EAR and do a dump in the local folder
        Iterator<URL> itResouces = null;
        try {
            itResouces = earDeployable.getArchive().getResources();
        } catch (EZBArchiveException e) {
            throw new EZBDeployerException("Cannot get the resources on the archive '" + earDeployable.getArchive() + "'.", e);
        }

        // Dump all resources
        List<File> entries = new ArrayList<File>();
        while (itResouces.hasNext()) {
            // Dump the input stream of the URL to the local folder
            URL url = itResouces.next();
            URLConnection urlConnection = null;
            try {
                urlConnection = url.openConnection();
            } catch (IOException e) {
                throw new EZBDeployerException("Cannot open the connection on the URL '" + url + "'.", e);
            }
            urlConnection.setDefaultUseCaches(false);
            InputStream is = null;
            try {
                is = urlConnection.getInputStream();
            } catch (IOException e) {
               throw new EZBDeployerException("Cannot get the input stream on the URL connection '" + urlConnection + "'.", e);
            }

            // unpack file
            File entryFile = new File(earUnpackedFolder, url.getPath().substring(url.getPath().lastIndexOf("!") + 2));
            logger.debug("Dumping url ''{0}'' inputstream into ''{1}''", url, entryFile);
            // Create directory (if some of them are missing)
            entryFile.getParentFile().mkdirs();

            // directory, not a file
            if (url.toString().endsWith("/")) {
                entryFile.mkdirs();
                continue;
            }

            // Dump the file
            try {
                FileUtils.dump(is, entryFile);
                entries.add(entryFile);
            } catch (FileUtilsException e) {
                throw new EZBDeployerException("Cannot dump the inputstream of url '" + url + "' into file '" + entryFile + "'.",
                        e);
            } finally {
                try {
                    is.close();
                } catch (IOException e) {
                    logger.warn("Problem when closing the input stream on url ''{0}''", url, e);
                }
            }
        }

        // Now, analyze each entry of the unpacked EAR
        unpackedDeployable = new EARDeployableImpl(builtArchive);
        unpackedDeployable.setOriginalDeployable(earDeployable);
        earDeployable.setUnpackedDeployable(unpackedDeployable);

        for (File entryFile : entries) {
            // Get archive
            EZBArchive tmpArchive = ArchiveManager.getInstance().getArchive(entryFile);

            // Valid archive, then get the type of the deployable
            if (tmpArchive != null) {
                // Analyze the new file
                EZBDeployable tmpDeployable = null;
                try {
                    tmpDeployable = DeployableHelper.getDeployable(tmpArchive);
                } catch (DeployableHelperException e) {
                    throw new EZBDeployerException("Cannot get a deployable on the file '" + entryFile + "'.", e);
                }
                // Add the deployable into the EAR deployable
                unpackedDeployable.addDeployable(tmpDeployable);
            }
        }

        return unpackedDeployable;
    }


}
