/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EasyBeansDeployer.java 1365 2007-05-09 16:41:15Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployer;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.easybeans.api.EZBContainer;
import org.objectweb.easybeans.api.EZBContainerException;
import org.objectweb.easybeans.deployer.api.EARDeployable;
import org.objectweb.easybeans.deployer.api.EJBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployer;
import org.objectweb.easybeans.deployer.api.EZBDeployerException;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.persistence.PersistenceUnitManager;

/**
 * Implementation of the Deployer for EasyBeans in standalone mode (or default
 * mode). It can deploy EJB-JAR or EAR deployable
 * @author Florent Benoit
 */
public class EasyBeansDeployer extends AbsDeployer implements EZBDeployer {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(EasyBeansDeployer.class);

    /**
     * Deploy a deployable. It can be an EJB jar, EAR, WAR, etc.
     * @param deployable a given deployable
     * @throws EZBDeployerException if the deployment is not done.
     */
    public void deploy(final EZBDeployable deployable) throws EZBDeployerException {
        checkSupportedDeployable(deployable);
        if (deployable instanceof EJBDeployable) {
            deployEJB((EJBDeployable) deployable);
        } else if (deployable instanceof EARDeployable) {
            // needs to unpack it before deploying it
            EARDeployable earDeployable = unpackEARDeployable((EARDeployable) deployable);
            deployEAR(earDeployable);
        }
    }

    /**
     * Undeploy the given deployable. It can be an EJB jar, EAR, WAR, etc.
     * @param deployable a given deployable to undeploy
     * @throws EZBDeployerException if the undeploy operation fails.
     */
    public void undeploy(final EZBDeployable deployable) throws EZBDeployerException {
        checkSupportedDeployable(deployable);
        if (deployable instanceof EJBDeployable) {
            throw new UnsupportedOperationException("Single EJB jar should not be removed by this deployer");
        } else if (deployable instanceof EARDeployable) {
            undeployEJB3FromEAR((EARDeployable) deployable);
        }



    }

    /**
     * Deploy an EAR (called by the deploy method).
     * @param earDeployable a given EAR deployable
     * @throws EZBDeployerException if the deployment is not done.
     */
    protected void deployEAR(final EARDeployable earDeployable) throws EZBDeployerException {

        // Needs to deploy all the EJB containers of the EAR with the same
        // classloader
        logger.info("Deploying {0}", earDeployable);

        // Get EJBs of this EAR
        List<EJBDeployable> ejbs = earDeployable.getEJBDeployables();

        // Build classloader
        ClassLoader appClassLoader = getClassLoaderForEAR(earDeployable);

        // Get Persistence unit manager
        PersistenceUnitManager persistenceUnitManager = getPersistenceUnitManager(earDeployable, appClassLoader);

        // Create containers for each EJB deployable
        List<EZBContainer> containers = new ArrayList<EZBContainer>();
        for (EJBDeployable ejb : ejbs) {
            containers.add(getEmbedded().createContainer(ejb.getArchive()));
        }

        // Configure containers
        for (EZBContainer container : containers) {
            // Set the classloader that needs to be used
            container.setClassLoader(appClassLoader);
            // Add persistence context found
            container.setPersistenceUnitManager(persistenceUnitManager);
        }

        // Start containers
        for (EZBContainer container : containers) {
            try {
                container.start();
            } catch (EZBContainerException e) {
                logger.error("Cannot start container {0}", container.getName(), e);
            }
        }

    }

    /**
     * Check that the given deployable is supported by this deployer. If it is
     * not supported, throw an error.
     * @param deployable the deployable that needs to be deployed
     * @throws EZBDeployerException if this deployable is not supported.
     */
    private void checkSupportedDeployable(final EZBDeployable deployable) throws EZBDeployerException {
        if (!(deployable instanceof EARDeployable || deployable instanceof EJBDeployable)) {
            throw new EZBDeployerException("The deployable '" + deployable + "' is not supported by this deployer");
        }
    }

}
